# CritHits by David Bowland
# ./addons/eventscripts/crithits/crithits.py

# >>> To configure this addon please see crithits.cfg (created when the addon is first loaded) <<<

"""
Provides TF2-style critical hits.
Requires only ES 2.1+
"""

from __future__ import division

import cfglib
import cmdlib
import es
import gamethread
import langlib
import playerlib
import random
import time
import usermsg
import weaponlib

import psyco
psyco.full()

from path import path


info = es.AddonInfo()
info.name     = 'CritHits'
info.version  = '1'
info.basename = 'crithits'
info.url      = 'http://addons.eventscripts.com/addons/view/' + info.basename
info.author   = 'SuperDave'

###

# Gib models and the min, max model
GIB_MODELS = {'models/gibs/hgibs.mdl': (1, 1),
              'models/gibs/hgibs_rib.mdl': (4, 10),
              'models/gibs/hgibs_scapula.mdl': (1, 2),
              'models/gibs/hgibs_spine.mdl': (1, 2)}


addonpath = path(es.getAddonPath(info.basename))

###

config = cfglib.AddonCFG(addonpath.joinpath(info.basename + '.cfg'))

config.text(info.name + ' release %s options' % info.version)
config.text('./addons/eventscripts/%(basename)s/%(basename)s.cfg' % {'basename': info.basename})
config.text(info.url)
config.text('')
config.text('Load this addon with: es_load ' + info.basename)
config.text('\n')
config.text('This addon turns on eventscripts_noisy. More information on eventscripts_noisy can be found here:')
config.text('http://www.eventscripts.com/pages/Eventscripts_noisy')
config.text('\n')

config.text('***** Base critical hit options *****')
cvar_base_percent          = config.cvar('crithits_base_percent',          2,   'Percent chance of inflicting critical damage [TF2 value: 2]')
cvar_base_percent_melee    = config.cvar('crithits_base_percent_melee',    15,  'Percent chance of inflicting critical damage with a melee weapon [TF2 value: 15]')
cvar_base_percent_endround = config.cvar('crithits_base_percent_endround', 100, 'Percent chance of inflicting critical damage for members of the winner team, after round end [TF2 value: 100]')
cvar_base_modifier_min     = config.cvar('crithits_base_modifier_min',     2.5, 'Minimum damage modifier for critical hits [TF2 value: presumed to be around 3.0]')
cvar_base_modifier_max     = config.cvar('crithits_base_modifier_max',     3.2, 'Maximum damage modifier for critical hits [TF2 value: presumed to be around 3.0]')
cvar_base_flashbang_max    = config.cvar('crithits_base_flashbang_max',    10,  'Maximum number of seconds a player can be blinded by a critical flashbang')
config.text('\n')

config.text('***** Rolling critical hit options *****')
cvar_rolling_maxpercent = config.cvar('crithits_rolling_maxpercent', 20,  'Maximum chance-to-crit percentage increase [TF2 value: 10]')
cvar_rolling_time       = config.cvar('crithits_rolling_time',       15,  'Number of seconds to retain damage information [TF2 value: 20]')
cvar_rolling_damage     = config.cvar('crithits_rolling_damage',     500, 'Amount of damage that will yeild the maximum chance-to-crit increase [TF2 value: 800]')
config.text('\n')

config.text('***** Rapid-fire critical hit options *****')
cvar_rapidfire_weapons    = config.cvar('crithits_rapidfire_exempt',    'knife;deagle;scout;awp;sg550;g3sg1;hegrenade;smokegrenade;flashbang', 'Weapons that are NOT rapid-fire (should have crit chance every shot), separated with ";"')
cvar_rapidfire_chancerate = config.cvar('crithits_rapidfire_chancerate', 1, 'Number of seconds between calculating critical hit chance on rapid-fire weapons [TF2 value: 1]')
cvar_rapidfire_duration   = config.cvar('crithits_rapidfire_duration',   2, 'Number of seconds rapid-fire weapons retain critical hit damage [TF2 value: 2]')
config.text('\n')

config.text('***** Announcement options *****')
cvar_announce_center = config.cvar('crithits_announce_center', 1, '1 = announce in center text, 0 = no center text announcements')
cvar_announce_echo   = config.cvar('crithits_announce_echo',   1, '1 = announce in console of attacker and victim, 2 = announce in console of all players, 0 = no console announcements')
config.text('\n')

config.text('***** Critical fire/hit sound options *****')
config.text('')
config.text('Use the following server command to add critical fire/hit sounds:')
config.text('')
config.text('crithits_sound_add <fire/hit> <"sound relative to ./sound/">')
config.text('\n')

config.text('Place your crithits_weapon_add commands below:')
config.command('crithits_sound_add')
config.text('\n')

config.text('Uncomment the following example options if you wish to use them:')
config.text('')
config.text('crithits_sound_add fire "weapons/stunstick/stunstick_fleshhit1.wav"')
config.text('crithits_sound_add fire "weapons/stunstick/stunstick_fleshhit2.wav"')
config.text('crithits_sound_add fire "weapons/stunstick/stunstick_impact1.wav"')
config.text('crithits_sound_add fire "weapons/stunstick/stunstick_impact2.wav"')
config.text('')
config.text('crithits_sound_add hit "npc/manhack/grind1.wav"')
config.text('crithits_sound_add hit "npc/manhack/grind2.wav"')
config.text('crithits_sound_add hit "npc/manhack/grind3.wav"')
config.text('crithits_sound_add hit "npc/manhack/grind4.wav"')
config.text('crithits_sound_add hit "npc/manhack/grind5.wav"')
config.text('\n')

config.text('Weapon-specific critical hit chance can be removed with the following server command:')
config.text('')
config.text('crithits_sound_remove <fire/hit> <"sound relative to ./sound/">')
config.text('')
config.text('Due to the fact custom critical chance can be added or removed removed above, this command will not be used by most users.')
config.text('\n')

config.text('***** Custom critical hit options *****')
config.text('')
config.text('Use the following server command to give a custom critical damage chance to weapons:')
config.text('')
config.text('crithits_weapon_add <weapon> <percent>')
config.text('')
config.text('See the bottom of these options for a list of acceptable weapon names.')
config.text('Using -1 as for the critical damage percent will remove critical hits for the weapon.')
config.text('\n')

config.text('Place your crithits_weapon_add commands below:')
config.command('crithits_weapon_add')
config.text('\n')

config.text('Uncomment the following example options if you wish to use them:')
config.text('')
config.text('crithits_weapon_add awp -1')
config.text('   No critical hits for the AWP')
config.text('crithits_weapon_add tmp 5')
config.text('   TMP starts with 5% chance of doing critical damage')
config.text('crithits_weapon_add mac10 5')
config.text('   MAC10 starts with 5% chance of doing critical damage')
config.text('\n')

config.text('Weapon-specific critical hit chance can be removed with the following server command:')
config.text('')
config.text('crithits_weapon_remove <weapon>')
config.text('')
config.text('Due to the fact custom critical chance can be added or removed removed above, this command will not be used by most users.')
config.text('\n')

config.text('***** Acceptable weapon names *****')
config.text('These are the weapon names acceptable for custom critical hit percentages along with their more common in-game names.')
config.text('This is just a reference--these are not options.')
config.text('')
config.text('deagle       - Night Hawk .50C')
config.text('ak47         - CV-47')
config.text('scout        - Schmidt Scout')
config.text('aug          - Bullpup')
config.text('g3sg1        - D3/AU1')
config.text('galil        - IDF Defender')
config.text('famas        - Clarion 5.56')
config.text('m4a1         - Maverick M4A1 Carbine')
config.text('sg552        - Krieg 552')
config.text('sg550        - Krieg 550 Commando')
config.text('m249         - M249')
config.text('awp          - Magnum Sniper Rifle')
config.text('tmp          - Schmidt Machine Pistol')
config.text('mp5navy      - KM Sub-Machine Gun')
config.text('glock        - 9x19mm Sidearm')
config.text('elite        - .40 Dual Elites')
config.text('m3           - Leone 12 Gauge Super')
config.text('xm1014       - Leone YG1265 Auto Shotgun')
config.text('usp          - KM .45 Tactical')
config.text('mac10        - Ingram MAC-10')
config.text('ump45        - KM UMP45')
config.text('p228         - 228 Compact')
config.text('fiveseven    - ES Five-Seven')
config.text('p90          - ES C90')
config.text('hegrenade    - HE Grenade')
config.text('flashbang    - Flashbang')
config.text('smokegrenade - Smoke Grenade')
config.text('knife        - Knife')

config.write()

###

ini = cfglib.AddonINI(addonpath.joinpath(info.basename + '_languages.ini'))

ini.setInitialComments([
 '# ./addons/eventscripts/%(basename)s/%(basename)s_languages.ini' % {'basename': info.basename},
 '# >>> For general addon configuration see ' + info.basename + '.cfg <<<',
 '',
 '# DO NOT translate words with $ or # in front!',
 '# For a list of language abbreviations please see ./addons/eventscripts/_libs/python/deflangs.ini'])

ini.addGroup('crit damage inflicted')
ini.addValueToGroup('crit damage inflicted', 'en', 'Critical damage inflicted')

ini.addGroup('crit damage taken')
ini.addValueToGroup('crit damage taken', 'en', 'Critical damage taken')

ini.addGroup('crit damage echo')
ini.addValueToGroup('crit damage echo', 'en', '$attacker inflicted critical damage on $victim with $weapon')

ini.write()


lang_text = langlib.Strings(ini)

###

class CritPlayer(object):
   """ Determines when a player should inflict critical damage """

   crittime   = None # Unix time critical hits should end for this player
   critweapon = None # Weapon critical hits apply to
   critcheck  = None # Unix time to next check for rapid-fire crits
   endround   = False # True if the round is over
   critmode   = False # True if the player's damage should be critical
   skipone    = False # True if we just executed critical damage (crits can't crit)

   # Last sounds
   lastfiresound = None
   lasthitsound  = None

   def __init__(self, userid):
      self.userid = userid
      self.recentdamage = {}

   ###

   def playerHurt(self, victim, weapon, damage):
      """ Retains damage information and inflicts critical hits """
      # If this damage was crit hit damage we ignore it
      if self.skipone:
         self.skipone = False
         return

      # Add the damage to our player's damage dictionary
      self.addDamage(weapon, damage)

      # If the player is in crit mode this is a critical hit
      if self.critmode:
         self.critHit(victim, weapon, damage)

   def weaponFire(self, weapon):
      """ Determines which shots should do critical damage """
      # If we don't recognize the weapon we exit
      weapon = weaponlib.getWeapon(weapon)
      if weapon is None:
         return

      # If we should crit, play the crit sound and add the effect
      if self.shouldCrit(weapon):
         self.critFire()
      # No crit
      else:
         self.critmode = False

   def flashbangDetonate(self, blinded_players):
      """ Detonates a critical flashbang """
      if self.critmode and cvar_base_flashbang_max > 0:
         self.critFlashbang(blinded_players)

   def smokeGrenadeDetonate(self, origin):
      """ Detonates a critical smoke grenade """
      if self.critmode:
         self.critSmokeGrenade(origin)

   def roundEnd(self):
      """ Changes the base critical chance for the player to the end-round percent """
      self.endround  = True
      self.critcheck = None

   ###

   def addDamage(self, weapon, damage):
      """ Adds the damage dealt to the dictionary for later recall """
      if weapon != self.critweapon:
         self.critweapon = weapon
         # If the weapon is not a rapid-fire weapons we give it a chance to crit every hit
         if weapon in map(weaponlib.getWeapon, str(cvar_rapidfire_weapons).split(';')):
            self.critcheck = None
         else:
            self.critcheck = 0

      now = time.time()
      self.recentdamage[now] = self.recentdamage.get(now, 0) + damage

   def shouldCrit(self, weapon):
      """
      Returns True if the player's shot should do critical damage,
      otherwise returns False
      """
      now = time.time()

      # If the player is in "crit time" then his or her shots will do critical damage
      if not self.crittime is None:
         if now <= self.crittime:
            return True
         self.crittime = None

      # Determine whether we should give another chance at a critical hit
      if self.critcheck is None or now >= self.critcheck:
         if not self.critcheck is None:
            self.critcheck = now + cvar_rapidfire_chancerate

         # Start with the base chance
         if self.endround:
            chance_to_crit = cvar_base_percent_endround

         elif weapon in specialweapons:
            chance_to_crit = specialweapons[weapon]
            # If the chance is negative crits are disabled for this weapon
            if chance_to_crit < 0:
               return False

         elif '#melee' in weapon.tags:
            chance_to_crit = cvar_base_percent_melee

         else:
            chance_to_crit = cvar_base_percent

         # Sum recent damage
         totaldamage = 0
         cutofftime  = now - cvar_rolling_time
         for x in list(self.recentdamage):
            if x < cutofftime:
               del self.recentdamage[x]
            else:
               totaldamage += self.recentdamage[x]

         # Add rolling damage percent
         chance_to_crit += min(totaldamage / cvar_rolling_damage, 1) * cvar_rolling_maxpercent

         # Finally, determine whether this is a crit
         if random.randint(1, 100) <= chance_to_crit:
            # If this is a rapid-fire crit check it will apply for the next few seconds
            if not self.critcheck is None:
               self.crittime = now + cvar_rapidfire_duration
            # This shot crits regardless of weapon
            return True

      # Sorry! No crit
      return False

   ###

   def critFire(self):
      """ Handles the sound and effect associated with firing a critical shot """
      # Play sound
      self.lastfiresound = sounds.getSound('fire', self.lastfiresound)
      es.playsound(self.userid, self.lastfiresound, 1)
      # Show effect
      es.effect('energysplash', ','.join(map(str, self.getChestLocation(self.userid))), '100,100,100', 1)
      # Enter crit mode
      self.critmode = True

      # Fire crit_fire event
      es.event('initialize', 'crit_fire')
      es.event('setint',     'crit_fire', 'userid', self.userid)
      es.event('setstring',  'crit_fire', 'weapon', str(self.critweapon).split('weapon_')[~0])
      es.event('fire',       'crit_fire')

   def critHit(self, victim, weapon, damage):
      """ Inflicts a critical hit """
      # Calculate damage
      damage = max(int(damage * self.getRandomModifier()), 1)
      # Inflict the damage
      self.damage(victim, damage, str(weapon))
      self.skipone = True

      # Play crit sound
      critvictim = players[victim]
      critvictim.lasthitsound = sounds.getSound('hit', critvictim.lasthitsound)
      es.playsound(victim, critvictim.lasthitsound, 1)

      # Show crit text
      if cvar_announce_center:
         es.centertell(victim, lang_text('crit damage taken', lang=playerlib.getPlayer(victim).getLanguage()))
         es.centertell(self.userid, lang_text('crit damage inflicted', lang=playerlib.getPlayer(self.userid).getLanguage()))

      tokens = {
       'victim': es.getplayername(victim),
       'attacker': es.getplayername(self.userid),
       'weapon': weapon,
       }
      if cvar_announce_echo == 2:
         for userid in es.getUseridList():
            usermsg.echo(userid, lang_text('crit damage echo', tokens, playerlib.getPlayer(userid).getLanguage()))
      elif cvar_announce_echo:
         usermsg.echo(victim, lang_text('crit damage echo', tokens, playerlib.getPlayer(victim).getLanguage()))
         usermsg.echo(self.userid, lang_text('crit damage echo', tokens, playerlib.getPlayer(self.userid).getLanguage()))

      # Show gib
      self.gib(victim)

      # Fire crit_hit event
      es.event('initialize', 'crit_hit')
      es.event('setint',     'crit_hit', 'userid',     victim)
      es.event('setint',     'crit_hit', 'attacker',   self.userid)
      es.event('setstring',  'crit_hit', 'weapon',     str(self.critweapon).split('weapon_')[~0])
      es.event('setint',     'crit_hit', 'dmg_health', damage)
      es.event('fire',       'crit_hit')

   def critFlashbang(self, blinded_players):
      """ Deals critical flashbang duration """
      # Critical flashbangs increase flash duration and always use 255 alpha
      modifier = self.getRandomModifier()
      for userid in blinded_players:
         player = playerlib.getPlayer(userid)
         duration = min(player.getFlashDuration() * modifier, cvar_base_flashbang_max)
         player.flash(255, duration)
         es.fadevolume(player, 100, 2, max(duration - 4, 1), 2)

   def critSmokeGrenade(self, origin):
      """ Creates smoke faster than normal smoke grenades for "critcal" smoke """
      # Show smoke immediately
      origin = ','.join(map(str, origin))
      es.effect('smoke', origin, 'smoke', 800, 0.1)
      for x in xrange(12):
         gamethread.delayedname(x * 0.25, 'crithits', es.effect, ('smoke', origin, 'smoke', 800, 0.1))

   ###

   def damage(self, victim, damage, weapon):
      """ Damages a player with point_hurt """
      # Give the player the targetname "critvictim"
      es.fire(victim, '!self', 'addoutput', 'targetname critvictim')

      # Create a point_hurt to damage the victim
      index = es.createentity('point_hurt') # Create the entity
      es.entitysetvalue(index, 'DamageTarget', 'critvictim') # Damage our victim
      es.entitysetvalue(index, 'DamageType', 32) # Default (fall) damage
      es.entitysetvalue(index, 'Damage', damage) # Damage to inflict
      es.entitysetvalue(index, 'targetname', 'crithits_hurt') # Give this entity a unique name
      es.entitysetvalue(index, 'classname', weapon) # Tell CS what type of weapon to credit
      es.spawnentity(index)

      # Hurt the victim
      es.fire(self.userid, 'crithits_hurt', 'hurt')
      # Remove the entity
      es.fire(self.userid, 'crithits_hurt', 'kill')

      # Give the victim a plain, meaningless targetname so he or she is not confused with future "critvictim" players
      es.fire(victim, '!self', 'addoutput', 'targetname none')

      # Add this damage to the dictionary of recent damage
      now = time.time()
      self.recentdamage[now] = self.recentdamage.get(now, 0) + damage

   def gib(self, victim):
      """ Shoots gibs from the victim to denote a critical hit """
      for x, model in enumerate(random.sample(GIB_MODELS, random.randint(1, len(GIB_MODELS)))):
         es.precachemodel(model)
         gamethread.delayedname(x, 'crithits', self._shootGibs, (victim, model, random.randint(*GIB_MODELS[model])))

   def _shootGibs(self, userid, model, count):
      """ Fires the env_shooter entity to make it shoot gibs """
      # If this is no longer a valid userid we cannot continue
      if not es.getuserid(userid) or playerlib.getPlayer(userid).isdead:
         return

      # Create the gibshooter entity and set its values
      index = es.createentity('env_shooter') # Create the gibshooter
      es.entitysetvalue(index, 'origin', ' '.join(map(str, self.getChestLocation(userid)))) # Location
      es.entitysetvalue(index, 'spawnflags', 5) # Repeatable, strict remove
      es.entitysetvalue(index, 'delay', 0) # Shoot all gibs at once
      es.entitysetvalue(index, 'gibangles', '0 90 0') # Shoot straight down
      es.entitysetvalue(index, 'm_flVelocity', 50) # Speed of gibs
      es.entitysetvalue(index, 'm_flVariance', 0.9) # Variance in gib direction
      es.entitysetvalue(index, 'm_flGibLife', 7) # TTL of gibs
      es.entitysetvalue(index, 'm_iGibs', random.randint(*GIB_MODELS[model])) # Number of gibs to shoot
      es.entitysetvalue(index, 'shootsounds', -1) # No shooting sounds
      es.entitysetvalue(index, 'simulation', 1) # Physics simulation
      es.entitysetvalue(index, 'gibanglevelocity', 25) # Angular spin of gibs
      es.entitysetvalue(index, 'model', model) # Gib model to use
      es.entitysetvalue(index, 'targetname', 'crithits_shooter') # Give this entity a unique name
      es.spawnentity(index)

      # Shoot the gibs
      es.fire(userid, 'crithits_shooter', 'Shoot')
      es.fire(userid, 'crithits_shooter', 'Kill')

   ###

   @staticmethod
   def getChestLocation(userid):
      """ Create an energy splash effect near a player's weapon """
      location = list(es.getplayerlocation(userid))
      location[2] = location[2] + es.getplayerprop(userid, 'CBasePlayer.localdata.m_vecViewOffset[2]') - 20
      return location

   @staticmethod
   def getRandomModifier():
      """ Returns a random modifier """
      return random.uniform(float(cvar_base_modifier_min), float(cvar_base_modifier_max)) - 1

###

class PlayerManager(dict):
   """ Stores CritPlayer instances for active players """

   def __getitem__(self, userid):
      userid = int(userid)
      if userid not in self:
         self[userid] = CritPlayer(userid)
      return super(PlayerManager, self).__getitem__(userid)

   def __delitem__(self, userid):
      userid = int(userid)
      if userid in self:
         super(PlayerManager, self).__delitem__(userid)

   def clear(self):
      gamethread.cancelDelayed('crithits')
      super(PlayerManager, self).clear()

players = PlayerManager()

###

class BlindedPlayerManager(set):
   """ Keeps track of which players have been blinded by each flashbang """

   def flashbangDetonate(self, attacker):
      """ Pass on the flashbang detonation then clear the set """
      players[attacker].flashbangDetonate(self)
      self.clear()

blindplayers = BlindedPlayerManager()

###

class WeaponCritManager(dict):
   """ Stores weapon-specific critical damage chances """

   def __init__(self):
      super(WeaponCritManager, self).__init__()
      cmdlib.registerServerCommand('crithits_weapon_add', self.addWeapon,
       'Syntax: crithits_weapon_add <weapon> <percent>\nGives a specific crit chance to a weapon, use -1 to disable crits')
      cmdlib.registerServerCommand('crithits_weapon_remove', self.removeWeapon,
       'Syntax: crithits_weapon_remove <weapon>\nRemoves any custom crit chance from a weapon')

   def addWeapon(self, args):
      """
      crithits_weapon_add <weapon> <percent>
      """
      if len(args) == 2:
         weapon = weaponlib.getWeapon(args[0])
         if weapon is None:
            es.dbgmsg(0, 'CritHits: Invalid weapon "%s"' % args[0])
            return
         self[weapon] = float(args[1])

      else:
         es.dbgmsg(0, 'Syntax: crithits_weapon_add <weapon> <percent>')

   def removeWeapon(self, args):
      """
      crithits_weapon_remove <weapon>
      """
      if len(args) == 1:
         weapon = weaponlib.getWeapon(args[0])
         if weapon in self:
            del self[weapon]

      else:
         es.dbgmsg(0, 'Syntax: crithits_weapon_remove <weapon>')

   @staticmethod
   def unregisterCommands():
      """ Unregister weapon-specific chance commands """
      cmdlib.unregisterServerCommand('crithits_weapon_add')
      cmdlib.unregisterServerCommand('crithits_weapon_remove')

specialweapons = WeaponCritManager()

###

class SoundManager(dict):
   """ Handles critical hit sounds """

   soundpath = addonpath.joinpath('sound') # Path to sound directory

   def __init__(self):
      super(SoundManager, self).__init__()

      self['fire'] = set()
      self['hit']  = set()

      cmdlib.registerServerCommand('crithits_sound_add', self.addSound,
       'Syntax: crithits_sound_add <fire/hit> <"path to sound from ./sound/">\nAdds a sound to the critical fire or critical hit sounds')
      cmdlib.registerServerCommand('crithits_sound_remove', self.removeSound,
       'Syntax: crithits_sound_remove <fire/hit> <"path to sound from ./sound/">\nRemoves a sound from the critical fire or critical hit sounds')

   def addSound(self, args):
      """
      crithits_sound_add <fire/hit> <"path to sound from ./sound/">
      """
      if len(args) == 2:
         soundtype = args[0].lower()
         if soundtype in self:
            self[soundtype].add(args[1])

         else:
            es.dbgmsg(0, 'CritHits: "%s" is not a valid sound type ("fire" or "hit" only)' % args[0])

      else:
         es.dbgmsg(0, 'Syntax: crithits_sound_add <fire/hit> <"path to sound from ./sound/">')

   def removeSound(self, args):
      """
      crithits_sound_remove <fire/hit> <"path to sound from ./sound/">
      """
      if len(args) == 2:
         soundtype = args[0].lower()
         if soundtype in self:
            if args[1] in self[soundtype]:
               self[soundtype].remove(args[1])

         else:
            es.dbgmsg(0, 'CritHits: "%s" is not a valid sound type ("fire" or "hit" only)' % args[0])

      else:
         es.dbgmsg(0, 'Syntax: crithits_sound_remove <fire/hit> <"path to sound from ./sound/">')

   def markForDownload(self):
      """ Marks critical hit sounds for download """
      for soundtype in self:
         for sound in self[soundtype]:
            if self.soundpath.joinpath(sound).isfile():
               es.stringtable('downloadables', sound)

   def getSound(self, soundtype, lastsound=None):
      """ Returns a random crit sound """
      choices = filter(lambda sound: sound != lastsound, self[soundtype])
      return random.choice(choices if choices else list(self[soundtype]))

   @staticmethod
   def unregisterCommands():
      """ Unregisters critical sound commands """
      cmdlib.unregisterServerCommand('crithits_sound_add')
      cmdlib.unregisterServerCommand('crithits_sound_remove')

sounds = SoundManager()

###

def loadEvents():
   """ Loads the .res file for the events """
   if addonpath.joinpath(info.basename + '.res').isfile():
      es.loadevents('declare', 'addons/eventscripts/%(basename)s/%(basename)s.res' % {'basename': info.basename})

   else:
      events = {'crit_fire': {'userid': 'short', 'weapon': 'string'},
                'crit_hit':  {'userid': 'short', 'attacker': 'short', 'weapon': 'string', 'dmg_health': 'byte'}}

      if not addonpath.joinpath('es_' + info.basename + '_events_db.txt').isfile():
         es.keygroupcreate(info.basename + '_events')
         for event_name in events:
            es.keycreate(info.basename + '_events', event_name)
            for event_var in events[event_name]:
               es.keysetvalue(info.basename + '_events', event_name, event_var, events[event_name][event_var])
         es.keygroupsave(info.basename + '_events', '|' + info.basename)
         es.keygroupdelete(info.basename + '_events')
      es.loadevents('declare', 'addons/eventscripts/' + info.basename + '/es_' + info.basename + '_events_db.txt')
      es.dbgmsg(0, info.name + ': Unable to load ' + info.basename + '.res! Please ensure it is in the ./' + info.basename + '/ directory.')

###

def load():
   config.execute()
   es.doblock('corelib\noisy_on')

   loadEvents()
   sounds.markForDownload()


def es_map_start(event_var):
   loadEvents()
   sounds.markForDownload()


def round_start(event_var):
   players.clear()
   blindplayers.clear()


def round_end(event_var):
   # No winner, no end crits
   if event_var['winner'] not in ('2', '3'):
      return
   winningteam = int(event_var['winner'])

   # Players on the winning team receive the end-round critical hit percent
   for userid in es.getUseridList():
      if es.getplayerteam(userid) == winningteam:
         players[userid].roundEnd()


def weapon_fire(event_var):
   players[event_var['userid']].weaponFire(event_var['weapon'])


def player_hurt(event_var):
   attacker = int(event_var['attacker'])
   if attacker:
      players[attacker].playerHurt(int(event_var['userid']), event_var['weapon'], int(event_var['dmg_health']))


def player_blind(event_var):
   blindplayers.add(int(event_var['userid']))


def flashbang_detonate(event_var):
   blindplayers.flashbangDetonate(event_var['userid'])


def smokegrenade_detonate(event_var):
   players[event_var['userid']].smokeGrenadeDetonate([float(event_var[x]) for x in 'xyz'])


def player_disconnect(event_var):
   del players[event_var['userid']]


def unload():
   es.doblock('corelib\noisy_off')

   players.clear()
   blindplayers.clear()

   specialweapons.unregisterCommands()
   sounds.unregisterCommands()

   for cvar in config.getCvars():
      es.ServerVar(cvar).set(0)